# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolic function (Functor)."""

import abc
import contextlib
import functools
import inspect
import threading
import types
import typing
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

from pyglove.core import object_utils
from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import base
from pyglove.core.symbolic import flags
from pyglove.core.symbolic import object as pg_object


class Functor(pg_object.Object, object_utils.Functor):
  """Symbolic functions (Functors).

  A symbolic function is a symbolic class with a ``__call__`` method, whose
  arguments can be bound partially, incrementally bound by attribute
  assignment, or provided at call time.

  Another useful trait is that a symbolic function is serializable, when
  its definition is imported by the target program and its arguments are also
  serializable. Therefore, it is very handy to move a symbolic function
  around in distributed scenarios.

  Symbolic functions can be created from regular function via
  :func:`pyglove.functor`::

    # Create a functor class using @pg.functor decorator.
    @pg.functor([
      ('a', pg.typing.Int(), 'Argument a'),
      # No field specification for 'b', which will be treated as any type.
    ])
    def sum(a, b=1, *args, **kwargs):
      return a + b + sum(args + kwargs.values())

    sum(1)()           # returns 2: prebind a=1, invoke with b=1 (default)
    sum(a=1)()         # returns 2: same as above.
    sum()(1)           # returns 2: bind a=1 at call time, b=1(default)

    sum(b=2)(1)        # returns 3: prebind b=2, invoke with a=1.
    sum(b=2)()         # wrong: `a` is not provided.

    sum(1)(2)                      # wrong: 'a' is provided multiple times.
    sum(1)(2, override_args=True)   # ok: override `a` value with 2.

    sum()(1, 2, 3, 4)  # returns 10: a=1, b=2, *args=[3, 4]
    sum(c=4)(1, 2, 3)  # returns 10: a=1, b=2, *args=[3], **kwargs={'c': 4}

  Or created by subclassing ``pg.Functor``::

    class Sum(pg.Functor):
      a: int
      b: int = 1

      def _call(self) -> int:
        return self.a + self.b

  Usage on subclassed functors is the same as functors created from functions.
  """

  # Allow assignment on symbolic attributes.
  allow_symbolic_assignment = True

  # Key for storing override members during call.
  _TLS_OVERRIDE_MEMBERS_KEY = '__override_members__'

  #
  # Customizable class traits.
  #

  @classmethod
  @property
  def is_subclassed_functor(cls) -> bool:
    """Returns True if this class is a subclassed Functor."""
    return cls.auto_schema

  @classmethod
  def _update_signatures_based_on_schema(cls):
    # Update the return value of subclassed functors.
    if cls.is_subclassed_functor:  # pylint: disable=using-constant-test
      private_call_signature = pg_typing.Signature.from_callable(
          cls._call, auto_typing=True, auto_doc=True,
      )
      if (
          len(private_call_signature.args) > 1
          or private_call_signature.kwonlyargs
      ):
        raise TypeError(
            '`_call` of a subclassed Functor should take no argument. '
            f'Encountered: {cls._call}.'
        )
      cls.__schema__.metadata['returns'] = private_call_signature.return_value

    # Update __init_ signature.
    init_signature = pg_typing.Signature.from_schema(
        cls.__schema__,
        name='__init__',
        module_name=cls.__module__,
        qualname=cls.__qualname__,
    )

    pseudo_init = init_signature.make_function(['pass'])

    # Save the original `Functor.__init__` before overriding it.
    if not hasattr(cls, '__orig_init__'):
      setattr(cls, '__orig_init__', cls.__init__)

    @object_utils.explicit_method_override
    @functools.wraps(pseudo_init)
    def _init(self, *args, **kwargs):
      self.__class__.__orig_init__(self, *args, **kwargs)

    setattr(cls, '__init__', _init)

    # Update __call__ signature.
    call_signature = pg_typing.Signature.from_schema(
        cls.__schema__,
        name='__call__',
        module_name=cls.__module__,
        qualname=cls.__qualname__,
        is_method=False,
    )
    setattr(cls, '__signature__', call_signature)

  def __new__(cls, *args, **kwargs):
    instance = object.__new__(cls)
    if flags.should_call_functors_during_init():
      instance.__init__(*args, **kwargs)
      return instance()
    return instance

  @object_utils.explicit_method_override
  def __init__(
      self,
      *args,
      root_path: Optional[object_utils.KeyPath] = None,
      override_args: bool = False,
      ignore_extra_args: bool = False,
      **kwargs):
    """Constructor.

    Args:
      *args: prebound positional arguments.
      root_path: The symbolic path for current object.
      override_args: If True, allows arguments provided during `__call__` to
        override existing bound arguments.
      ignore_extra_args: If True, unsupported arguments can be passed in
        during `__call__` without using them. Otherwise, calling with
        unsupported arguments will raise error.
      **kwargs: prebound keyword arguments.

    Raises:
      KeyError: constructor got unexpected arguments.
    """
    # NOTE(daiyip): Since Functor is usually late bound (until call time),
    # we pass `allow_partial=True` during functor construction.
    _ = kwargs.pop('allow_partial', None)

    varargs = None
    signature = self.__signature__
    if len(args) > len(signature.args):
      if signature.varargs:
        varargs = list(args[len(signature.args) :])
        args = args[: len(signature.args)]
      else:
        arg_phrase = object_utils.auto_plural(len(signature.args), 'argument')
        was_phrase = object_utils.auto_plural(len(args), 'was', 'were')
        raise TypeError(
            f'{signature.id}() takes {len(signature.args)} '
            f'positional {arg_phrase} but {len(args)} {was_phrase} given.'
        )

    bound_kwargs = dict()
    for i, v in enumerate(args):
      if pg_typing.MISSING_VALUE != v:
        bound_kwargs[signature.args[i].name] = v

    if varargs is not None:
      bound_kwargs[signature.varargs.name] = varargs

    for k, v in kwargs.items():
      if pg_typing.MISSING_VALUE != v:
        if k in bound_kwargs:
          raise TypeError(
              f'{signature.id}() got multiple values for keyword '
              f'argument {k!r}.'
          )
        bound_kwargs[k] = v

    default_args = set()
    non_default_args = set(bound_kwargs)

    for arg_spec in signature.named_args:
      if not arg_spec.value_spec.has_default:
        continue
      arg_name = arg_spec.name
      if arg_name not in non_default_args:
        default_args.add(arg_name)
      elif bound_kwargs[arg_name] == arg_spec.value_spec.default:
        default_args.add(arg_name)
        non_default_args.discard(arg_name)

    if signature.varargs and not varargs:
      default_args.add(signature.varargs.name)

    super().__init__(allow_partial=True,
                     root_path=root_path,
                     **bound_kwargs)

    self._non_default_args = non_default_args
    self._default_args = default_args
    self._specified_args = set(bound_kwargs)
    self._override_args = override_args
    self._ignore_extra_args = ignore_extra_args

    # For subclassed Functor, we use thread-local storage for storing temporary
    # member overrides from the arguments during functor call.
    self._tls = threading.local() if self.is_subclassed_functor else None

  def _sym_inferred(self, key: str, **kwargs: Any) -> Any:
    """Overrides method to allow member overrides during call."""
    if self._tls is not None:
      overrides = getattr(self._tls, Functor._TLS_OVERRIDE_MEMBERS_KEY, {})
      v = overrides.get(key, pg_typing.MISSING_VALUE)
      if pg_typing.MISSING_VALUE != v:
        return overrides[key]
    return super()._sym_inferred(key, **kwargs)

  def _sym_clone(self, deep: bool, memo: Any = None) -> 'Functor':
    """Override to copy bound args."""
    other = super()._sym_clone(deep, memo)
    # pylint: disable=protected-access
    other._non_default_args = set(self._non_default_args)
    other._default_args = self._default_args
    other._specified_args = self._specified_args
    other._override_args = self._override_args
    other._ignore_extra_args = self._ignore_extra_args
    # pylint: enable=protected-access
    return typing.cast(Functor, other)

  def _on_change(
      self, field_updates: Dict[object_utils.KeyPath, base.FieldUpdate]):
    """Custom handling field change to update bound args."""
    for relative_path, update in field_updates.items():
      assert relative_path
      if len(relative_path) != 1:
        continue
      arg_name = str(relative_path)
      if update.field.default_value == update.new_value:
        if update.field.value.has_default:
          self._default_args.add(arg_name)
        self._non_default_args.discard(arg_name)
      else:
        self._default_args.discard(arg_name)
        self._non_default_args.add(arg_name)

      if update.new_value == pg_typing.MISSING_VALUE:
        self._specified_args.discard(arg_name)
      else:
        self._specified_args.add(arg_name)

  def __delattr__(self, name: str) -> None:
    """Discard a previously bound argument and reset to its default value."""
    del self._sym_attributes[name]
    if self.__signature__.get_value_spec(name).has_default:
      self._default_args.add(name)
    self._specified_args.discard(name)
    self._non_default_args.discard(name)

  def _sym_missing(self) -> Dict[str, Any]:
    """Returns missing values for Functor.

    Semantically unbound arguments are not missing, thus we only return partial
    bound arguments in `sym_missing`. As a result, a functor is partial only
    when any of its bound arguments is partial.

    Returns:
      A dict of missing key (or path) to missing value.
    """
    missing = dict()
    for k, v in self._sym_attributes.items():
      if pg_typing.MISSING_VALUE != v and isinstance(v, base.Symbolic):
        missing_child = v.sym_missing(flatten=False)
        if missing_child:
          missing[k] = missing_child
    return missing

  @property
  def specified_args(self) -> Set[str]:
    """Returns user specified argument names."""
    return self._specified_args

  @property
  def non_default_args(self) -> Set[str]:
    """Returns the names of bound arguments whose values are not the default."""
    return self._non_default_args

  @property
  def default_args(self) -> Set[str]:
    """Returns the names of bound argument whose values are the default."""
    return self._default_args

  @property
  def bound_args(self) -> Set[str]:
    """Returns bound argument names."""
    return self._non_default_args | self._default_args

  @property
  def unbound_args(self) -> Set[str]:
    """Returns unbound argument names."""
    return set([name for name in self._sym_attributes.keys()
                if name not in self.bound_args])

  @property
  def is_fully_bound(self) -> bool:
    """Returns if all arguments of functor is bound."""
    return (len(self._non_default_args) + len(self._default_args)
            == len(self._sym_attributes))

  def _call(self, *args, **kwargs) -> Callable:  # pylint: disable=g-bare-generic
    """Actual function logic. Subclasses should implement this method."""
    raise NotImplementedError()

  # TODO(b/183649930): We pretend that _call is not abstract to avoid
  # [not-instantiable] errors from pytype.
  if not typing.TYPE_CHECKING:
    _call = abc.abstractmethod(_call)

  def __call__(self, *args, **kwargs) -> Any:
    """Call with late bound arguments.

    Args:
      *args: list arguments.
      **kwargs: keyword arguments.

    Returns:
      Any.

    Raises:
      TypeError: got multiple values for arguments or extra argument name.
    """
    args, kwargs = self._parse_call_time_overrides(*args, **kwargs)
    signature = self.__signature__

    if self.is_subclassed_functor:
      for arg_spec, arg_value in zip(signature.args, args):
        kwargs[arg_spec.name] = arg_value

      # Temporarily override members with argument values from the call.
      with self._apply_call_time_overrides_to_members(**kwargs):
        return_value = self._call()
    else:
      return_value = self._call(*args, **kwargs)

    # Return value check.
    if (
        signature.return_value
        and flags.is_type_check_enabled()
        and pg_typing.MISSING_VALUE != return_value
    ):
      return_value = signature.return_value.apply(
          return_value, root_path=self.sym_path + 'returns'
      )
    if flags.is_tracking_origin() and isinstance(return_value, base.Symbolic):
      return_value.sym_setorigin(self, 'return')
    return return_value

  @contextlib.contextmanager
  def _apply_call_time_overrides_to_members(self, **kwargs):
    """Overrides member values within the scope."""
    assert self._tls is not None
    setattr(self._tls, Functor._TLS_OVERRIDE_MEMBERS_KEY, kwargs)
    try:
      yield
    finally:
      delattr(self._tls, Functor._TLS_OVERRIDE_MEMBERS_KEY)

  def _parse_call_time_overrides(
      self, *args, **kwargs
  ) -> Tuple[List[Any], Dict[str, Any]]:
    """Parses positional and keyword arguments from call-time overrides."""
    override_args = kwargs.pop('override_args', self._override_args)
    ignore_extra_args = kwargs.pop('ignore_extra_args', self._ignore_extra_args)

    signature = self.__signature__
    if len(args) > len(signature.args) and not signature.has_varargs:
      if ignore_extra_args:
        args = args[: len(signature.args)]
      else:
        arg_phrase = object_utils.auto_plural(len(signature.args), 'argument')
        was_phrase = object_utils.auto_plural(len(args), 'was', 'were')
        raise TypeError(
            f'{signature.id}() takes {len(signature.args)} '
            f'positional {arg_phrase} but {len(args)} {was_phrase} given.'
        )

    keyword_args = {
        k: v for k, v in self._sym_attributes.items()
        if k in self._specified_args
    }
    assert len(keyword_args) == len(self._specified_args)

    # Work out varargs when positional arguments are provided.
    varargs = None
    if signature.has_varargs:
      varargs = list(args[len(signature.args) :])
      if flags.is_type_check_enabled():
        varargs = [
            signature.varargs.value_spec.element.value.apply(
                v, root_path=self.sym_path + signature.varargs.name
            )
            for v in varargs
        ]
      args = args[: len(signature.args)]

    # Convert positional arguments to keyword arguments so we can map them back
    # later.
    for i in range(len(args)):
      arg_spec = signature.args[i]
      arg_name = arg_spec.name
      if arg_name in self._specified_args:
        if not override_args:
          raise TypeError(
              f'{signature.id}() got new value for argument {arg_name!r} '
              f"from position {i}, but 'override_args' is set to False. "
              f'Old value: {keyword_args[arg_name]!r}, new value: {args[i]!r}.'
          )
      arg_value = args[i]
      if flags.is_type_check_enabled():
        arg_value = arg_spec.value_spec.apply(
            arg_value, root_path=self.sym_path + arg_name)
      keyword_args[arg_name] = arg_value

    for arg_name, arg_value in kwargs.items():
      if arg_name in self._specified_args:
        if not override_args:
          raise TypeError(
              f'{signature.id}() got new value for argument {arg_name!r} '
              "from keyword argument, while 'override_args' is set to "
              f'False. Old value: {keyword_args[arg_name]!r}, '
              f'new value: {arg_value!r}.'
          )
      arg_spec = signature.get_value_spec(arg_name)
      if arg_spec and flags.is_type_check_enabled():
        arg_value = arg_spec.apply(
            arg_value, root_path=self.sym_path + arg_name)
        keyword_args[arg_name] = arg_value
      elif not ignore_extra_args:
        raise TypeError(
            f'{signature.id}() got an unexpected keyword argument {arg_name!r}.'
        )

    # Use positional arguments if possible. This allows us to handle varargs
    # with simplicity.
    list_args = []
    missing_required_arg_names = []
    for arg in signature.args:
      if arg.name in keyword_args:
        list_args.append(keyword_args[arg.name])
        del keyword_args[arg.name]
      elif arg.value_spec.default != pg_typing.MISSING_VALUE:
        list_args.append(arg.value_spec.default)
      else:
        missing_required_arg_names.append(arg.name)

    if missing_required_arg_names:
      arg_phrase = object_utils.auto_plural(
          len(missing_required_arg_names), 'argument')
      args_str = object_utils.comma_delimited_str(missing_required_arg_names)
      raise TypeError(
          f'{signature.id}() missing {len(missing_required_arg_names)} '
          f'required positional {arg_phrase}: {args_str}.'
      )

    if signature.has_varargs:
      prebound_varargs = keyword_args.pop(signature.varargs.name, None)
      varargs = varargs or prebound_varargs
      if varargs:
        list_args.extend(varargs)
    return list_args, keyword_args


def functor(
    args: Optional[List[Union[
        Tuple[Union[str, pg_typing.KeySpec], pg_typing.ValueSpec, str],
        Tuple[Union[str, pg_typing.KeySpec], pg_typing.ValueSpec, str, Any]]]
    ] = None,    # pylint: disable=bad-continuation
    returns: Optional[pg_typing.ValueSpec] = None,
    base_class: Optional[Type[Functor]] = None,
    **kwargs):
  """Function/Decorator for creating symbolic function from regular function.

  Example::

    # Create a symbolic function without specifying the
    # validation rules for arguments.
    @pg.functor
    def foo(x, y):
      return x + y

    f = foo(1, 2)
    assert f() == 3

    # Create a symbolic function with specifying the
    # the validation rules for argument 'a', 'args', and 'kwargs'.
    @pg.functor([
      ('a', pg.typing.Int()),
      ('b', pg.typing.Float()),
      ('args', pg.List(pg.typing.Int())),
      (pg.typing.StrKey(), pg.typing.Int())
    ])
    def bar(a, b, c, *args, **kwargs):
      return a * b / c + sum(args) + sum(kwargs.values())

  See :class:`pyglove.Functor` for more details on symbolic function.

  Args:
    args: A list of tuples that defines the schema for function arguments.
      Please see `functor_class` for detailed explanation of `args`.
    returns: Optional value spec for return value.
    base_class: Optional base class derived from `symbolic.Functor`. If None,
      returning functor will inherit from `symbolic.Functor`.
    **kwargs: Keyword arguments for infrequently used options: Acceptable
      keywords are:  * `serialization_key`: An optional string to be used as the
      serialization key for the class during `sym_jsonify`. If None,
      `cls.__type_name__` will be used. This is introduced for scenarios when we
      want to relocate a class, before the downstream can recognize the new
      location, we need the class to serialize it using previous key. *
      `additional_keys`: An optional list of strings as additional keys to
      deserialize an object of the registered class. This can be useful when we
      need to relocate or rename the registered class while being able to load
      existing serialized JSON values.

  Returns:
    A function that converts a regular function into a symbolic function.
  """
  if inspect.isfunction(args):
    assert returns is None
    return functor_class(
        typing.cast(Callable[..., Any], args),
        base_class=base_class,
        add_to_registry=True,
        **kwargs,
    )
  return lambda fn: functor_class(  # pylint: disable=g-long-lambda  # pytype: disable=wrong-arg-types
      fn, args, returns,
      base_class=base_class,
      add_to_registry=True,
      **kwargs)


def functor_class(
    func: types.FunctionType,
    args: Union[
        List[Union[pg_typing.Field, pg_typing.FieldDef]],
        Dict[pg_typing.FieldKeyDef, pg_typing.FieldValueDef],
        None
    ] = None,
    returns: Optional[pg_typing.ValueSpec] = None,
    base_class: Optional[Type[Functor]] = None,
    *,
    auto_doc: bool = False,
    auto_typing: bool = False,
    serialization_key: Optional[str] = None,
    additional_keys: Optional[List[str]] = None,
    add_to_registry: bool = False,
) -> Type[Functor]:
  """Returns a functor class from a function.

  Args:
    func: Function to be wrapped into a functor.
    args: Symbolic args specification. `args` is a list of tuples, each
      describes an argument from the input function. Each tuple is the format of
      (<argumment-name>, <value-spec>, [description], [metadata-objects]).
      `argument-name` - a `str` or `pg_typing.StrKey` object. When
      `pg_typing.StrKey` is used, it describes the wildcard keyword argument.
      `value-spec` - a `pg_typing.ValueSpec` object or equivalent, e.g.
      primitive values which will be converted to ValueSpec implementation
      according to its type and used as its default value. `description` - a
      string to describe the agument. `metadata-objects` - an optional list of
      any type, which can be used to generate code according to the schema.
      There are notable rules in filling the `args`: 1) When `args` is None or
      arguments from the function signature are missing from it, `schema.Field`
      for these fields will be automatically generated and inserted into `args`.
      That being said, every arguments in input function will have a
      `schema.Field` counterpart in `Functor.__schema__.fields` sorted by the
      declaration order of each argument in the function signature ( other than
      the order in `args`). 2) Default argument values are specified along with
      function definition as regular python functions, instead of being set at
      `schema.Field` level. But validation rules can be set using `args` and
      apply to argument values.
      For example::

        @pg.functor([('c', pg.typing.Int(min_value=0), 'Arg c')])
          def foo(a, b, c=1, **kwargs):
            return a + b + c + sum(kwargs.values())

          assert foo.schema.fields() == [
              pg.typing.Field('a', pg.Any(), 'Argument a'.),
              pg.typing.Field('b', pg.Any(), 'Argument b'.),
              pg.typing.Field('c', pg.typing.Int(), 'Arg c.),
              pg.typing.Filed(
                  pg.typing.StrKey(), pg.Any(), 'Other arguments.')
          ]
          # Prebind a=1, b=2, with default value c=1.
          assert foo(1, 2)() == 4
    returns: Optional schema specification for the return value.
    base_class: Optional base class (derived from `symbolic.Functor`). If None,
      returned type will inherit from `Functor` directly.
    auto_doc: If True, the descriptions of argument fields will be inherited
      from funciton docstr if they are not explicitly specified through
      ``args``.
    auto_typing: If True, the value spec for constraining each argument will be
      inferred from its annotation. Otherwise the value specs for all arguments
      will be ``pg.typing.Any()``.
    serialization_key: An optional string to be used as the serialization key
      for the class during `sym_jsonify`. If None, `cls.__type_name__` will be
      used. This is introduced for scenarios when we want to relocate a class,
      before the downstream can recognize the new location, we need the class to
      serialize it using previous key.
    additional_keys: An optional list of strings as additional keys to
      deserialize an object of the registered class. This can be useful when we
      need to relocate or rename the registered class while being able to load
      existing serialized JSON values.
    add_to_registry: If True, the newly created functor class will be added to
      the registry for deserialization.

  Returns:
    `symbolic.Functor` subclass that wraps input function.

  Raises:
    KeyError: names of symbolic arguments are not compatible with
      function signature.
    TypeError: types of symbolic arguments are not compatible with
      function signature.
    ValueError: default values of symbolic arguments are not compatible
      with  function signature.
  """
  if not inspect.isfunction(func):
    raise TypeError(f'{func!r} is not a function.')

  class _Functor(base_class or Functor):
    """Functor wrapper for input function."""

    # The schema for function-based Functor will be inferred from the function
    # signature. Therefore we do not infer the schema automatically during class
    # creation.
    auto_schema = False

    # Do not infer symbolic fields from annotations, since this functor is
    # created from function definition which does not have class-level
    # attributes.
    infer_symbolic_fields_from_annotations = True

    def _call(self, *args, **kwargs):
      return func(*args, **kwargs)

  cls = typing.cast(Type[Functor], _Functor)
  cls.__name__ = func.__name__
  cls.__qualname__ = func.__qualname__
  cls.__module__ = getattr(func, '__module__', 'wrapper')
  cls.__doc__ = func.__doc__

  # Enable automatic registration for subclass.
  cls.auto_register = True

  # Apply function schema.
  cls.apply_schema(
      pg_typing.schema(
          func, args, returns, auto_doc=auto_doc, auto_typing=auto_typing
      )
  )

  # Register functor class for deserialization if needed.
  if add_to_registry:
    cls.register_for_deserialization(serialization_key, additional_keys)
  return cls


def as_functor(
    func: Callable,  # pylint: disable=g-bare-generic
    ignore_extra_args: bool = False) -> Functor:
  """Make a functor object from a regular python function.

  NOTE(daiyip): This method is designed to create on-the-go functor object,
  usually for lambdas. To create a reusable functor class, please use
  `functor_class` method.

  Args:
    func: A regular python function.
    ignore_extra_args: If True, extra argument which is not acceptable by `func`
      will be ignored.

  Returns:
    Functor object from input function.
  """
  return functor_class(func)(ignore_extra_args=ignore_extra_args)  # pytype: disable=not-instantiable



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface of PyGlove's view system: Content, View, and Extension.

PyGlove introduces a flexible and extensible view system for rendering objects
in different formats, such as text or HTML. This system allows users to plug in
various predefined views or create their own custom views with minimal effort.
The two core concepts of this system are ``View`` and ``Extension``.

A ``View`` defines how objects are displayed, with specific classes like
``HtmlView`` and more specialized subclasses such as ``HtmlTreeView``. Each
concrete ``View`` class has a ``VIEW_ID`` that can be used to select the view
when calling functions like ``pg.view`` or ``pg.to_html``. Views provide default
rendering behavior for built-in Python types such as ``int``, ``str``, ``bool``,
``list``, and ``dict``. However, users can extend this behavior by subclassing
the ``View.Extension`` class to define custom rendering logic for user-defined
classes.

.. code-block::

     pg.to_html
         |
         | (calls)
         v
      pg.view
         |
         | (refers)
         v
       pg.View --------------------------------> pg.View.Extension
         ^                   (calls)                     ^
         | (extends)                                     |
    +-----------+                                        |
    |           |                                        |
   ...   pg.views.HtmlView   ---------------------> pg.views.HtmlView.Extension
                ^             (calls)                    ^
                | (extends)                              |  (extends)
    pg.views.HtmlTreeView  ------------------> pg.views.HtmlTreeView.Extension
                              (calls)                 ^
                                                      |
                                                  pg.Symbolic
                                                      ^
                                                      |
                                                   pg.Object

Each ``View`` class also defines an inner ``Extension`` class that manages
interactions with user-defined types. The ``Extension`` class typically
implements several methods chosen by the ``View`` developer, such as
``_html_tree_view_render``, ``_html_tree_view__summary``, etc., depending on
the specific renderin needs.
The ``View.extension_method(<extension_method_name>)`` decorator enables
flexible mapping between a ``View`` method and its corresponding ``Extension``
method. This allows users to define custom rendering logic within their
``Extension`` class and bind it to the view.

An example of introducing a new view is described below:

.. code-block:: python

  class MyView(pg.View):
    VIEW_ID = 'my_view'

    class Extension(pg.View.Extension):
      def _myview_render(self, **kwargs):
        return ...

      def _myview_part1(self, **kwargs):
        return ...

      def _myview_part2(self, **kwargs):
        return ...

    @pg.View.extension_method('_myview_render')
    def render(self, value: Any, **kwargs):
      part1 = self.render_part1(value.part1, **kwargs)
      part2 = self.render_part2(value.part2, **kwargs)
      return ...

    @pg.View.extension_method('_myview_part1')
    def render_part1(self, value: Any, **kwargs):
      return ...

    @pg.View.extension_method('_myview_part2')
    def render_part2(self, value: Any, **kwargs):
      return ...

The ``View.extension_method`` decorator also supports multiple views for a
single user class by allowing multi-inheritance of ``Extension`` classes.
For example, a class can support both ``View1`` and ``View2`` by inheriting
their respective ``Extension`` classes and implementing methods like
``_render_view1`` and ``_render_view2`` to handle different view-specific
rendering:

.. code-block:: python

  class MyObject(MyView1.Extension, MyView2.Extension):
    def _myview1_render(self, value, **kwargs):
      return ...

    def _myview2_render(self, value, **kwargs):
      return ...

For ``extension_method``-decorated ``View`` methods, they must include a
``value`` argument, which is the target object to be rendered. The view will
delegate rendering to the user logic only when the ``value`` is an instance of
the ``Extension`` class.

`pg.view` is the function to display an object with a specific view. It takes
a ``view_id`` argument to specify which view to use, and returns a ``Content``
object, which acts as the media to be displayed. For example::

.. code-block:: python

  pg.view([1, 2, MyObject(...)], view_id='my_view1')

"""
import abc
import collections
import contextlib
import copy as copy_lib
import functools
import inspect
import io
import os
import types
from typing import Any, Callable, Dict, Iterator, Optional, Sequence, Set, Type, Union

from pyglove.core import io as pg_io
from pyglove.core import object_utils
from pyglove.core import typing as pg_typing


# Type definition for the value filter function.
NodeFilter = Callable[
    [
        object_utils.KeyPath,    # The path to the value.
        Any,        # Current value.
        Any,        # Parent value
    ],
    bool            # Whether to include the value.
]


_VIEW_ARGS_PRESET_NAME = 'pyglove_view_args'
_VIEW_REGISTRY = {}
_TLS_KEY_OPERAND_STACK_BY_METHOD = '__view_operand_stack__'
_TLS_KEY_VIEW_OPTIONS = '__view_options__'


class Content(object_utils.Formattable, metaclass=abc.ABCMeta):
  """Content: A type of media to be displayed in a view.

  For example, `pg.Html` is a `Content` type that represents HTML to be
  displayed in an HTML view.
  """

  WritableTypes = Union[    # pylint: disable=invalid-name
      str,
      'Content',
      Callable[[], Union[str, 'Content', None]],
      None
  ]

  class SharedParts(object_utils.Formattable):
    """A part of the content that should appear just once.

    For example, `pg.Html.Styles` is a `SharedParts` type that represents
    a group of CSS styles that should only be included once in the HEAD section.
    """

    __slots__ = ('_parts',)

    def __init__(
        self,
        *parts: Union[str, 'Content.SharedParts', None]
    ) -> None:
      self._parts: Dict[str, int] = collections.defaultdict(int)
      self.add(*parts)

    def add(self, *parts: Union[str, 'Content.SharedParts', None]) -> bool:
      """Adds one or multiple parts."""
      updated = False
      for part in parts:
        if part is None:
          continue
        if isinstance(part, Content.SharedParts):
          assert isinstance(part, self.__class__), (part, self.__class__)
          for p, count in part.parts.items():
            self._parts[p] += count
            updated = True
        else:
          assert isinstance(part, str), part
          updated |= (part not in self._parts)
          self._parts[part] += 1

      if updated:
        self.__dict__.pop('content', None)
      return updated

    @property
    def parts(self) -> Dict[str, int]:
      """Returns all parts and their reference counts."""
      return self._parts

    def __bool__(self) -> bool:
      """Returns True if there is any part."""
      return bool(self._parts)

    def __contains__(self, part: str) -> bool:
      """Returns True if the part is in the shared parts."""
      return part in self._parts

    def __iter__(self) -> Iterator[str]:
      """Iterates all parts."""
      return iter(self._parts.keys())

    def __copy__(self) -> 'Content.SharedParts':
      """Returns a copy of the shared parts."""
      return self.__class__(self)   # pytype: disable=not-instantiable

    def __eq__(self, other: Any):
      if not isinstance(other, self.__class__):
        return False
      return set(self._parts.keys()) == set(other.parts.keys())

    def __ne__(self, other: Any):
      return not self.__eq__(other)

    def format(
        self,
        compact: bool = False,
        verbose: bool = True,
        root_indent: int = 0,
        **kwargs
    ) -> str:
      if compact:
        return object_utils.kvlist_str(
            [
                ('parts', self._parts, {}),
            ],
            label=self.__class__.__name__,
            compact=compact,
            verbose=verbose,
            root_indent=root_indent,
            bracket_type=object_utils.BracketType.ROUND,
        )
      return self.content

    @property
    @abc.abstractmethod
    def content(self) -> str:
      """Returns the content string representing the the shared parts."""

  __slots__ = ('_content_stream', '_shared_parts',)

  def __init__(
      self,
      *content: WritableTypes,
      **shared_parts: 'Content.SharedParts'
  ):
    self._content_stream = io.StringIO()
    self._shared_parts = shared_parts

    for c in content:
      c = self._to_content(c)
      if c is None:
        continue
      elif isinstance(c, str):
        self._content_stream.write(c)
      else:
        self.write(c)

  @functools.cached_property
  def content(self) -> str:
    """Returns the content."""
    return self._content_stream.getvalue()

  @property
  def shared_parts(self) -> Dict[str, 'Content.SharedParts']:
    """Returns the shared parts."""
    return self._shared_parts

  def write(
      self, *parts: WritableTypes, shared_parts_only: bool = False
  ) -> 'Content':
    """Writes one or more parts to current Content.

    Args:
      *parts: The parts to be written. Each part can be a string, a Content
        object, a callable that returns one of the above, or None.
      shared_parts_only: If True, only write the shared parts.

    Returns:
      The current Content object for chaining.
    """
    content_updated = False
    for p in parts:
      p = self._to_content(p)
      if p is None:
        continue

      if not isinstance(p, (str, self.__class__)):
        raise TypeError(
            f'{p!r} ({type(p)}) cannot be writable. '
            f'Only str, None, {self.__class__.__name__} and callable object '
            'that returns one of them are supported.'
        )

      if isinstance(p, Content):
        current = self._shared_parts
        for k, v in p.shared_parts.items():
          # Since `p` is the same type of `self`, we expect they have the
          # same set of shared parts, which is determined at the __init__ time.
          current[k].add(v)
        p = p.content

      if not shared_parts_only:
        self._content_stream.write(p)
        content_updated = True

    if content_updated:
      self.__dict__.pop('content', None)
    return self

  def save(self, file: str, **kwargs):
    """Save content to a file."""
    pg_io.mkdirs(os.path.dirname(file), exist_ok=True)
    pg_io.writefile(file, self.to_str(**kwargs))

  def __add__(self, other: WritableTypes) -> 'Content':
    """Operator +: Concatenates two Content objects."""
    other = self._to_content(other)
    if not other:
      return self
    s = copy_lib.deepcopy(self)
    s.write(other)
    return s

  def __radd__(self, other: WritableTypes) -> 'Content':
    """Right-hand operator +: concatenates two Content objects."""
    s = self.from_value(other, copy=True)
    if s is None:
      return self
    s.write(self)
    return s

  def format(self,
             compact: bool = False,
             verbose: bool = True,
             root_indent: int = 0,
             content_only: bool = False,
             **kwargs) -> str:
    """Formats the Content object."""
    del kwargs
    if compact:
      return object_utils.kvlist_str(
          [
              ('content', self.content, ''),
          ] + [
              (k, v, None) for k, v in self._shared_parts.items()
          ],
          label=self.__class__.__name__,
          compact=compact,
          verbose=verbose,
          root_indent=root_indent,
          bracket_type=object_utils.BracketType.ROUND,
      )
    return self.to_str(content_only=content_only)

  @abc.abstractmethod
  def to_str(self, *, content_only: bool = False, **kwargs) -> str:
    """Returns the string representation of the content."""

  def __eq__(self, other: Any) -> bool:
    if not isinstance(other, self.__class__):
      return False
    return (
        self.content == other.content
        and self._shared_parts == other._shared_parts  # pylint: disable=protected-access
    )

  def __ne__(self, other: Any) -> bool:
    return not self.__eq__(other)

  def __hash__(self):
    return hash(self.to_str())

  @classmethod
  def from_value(
      cls,
      value: WritableTypes,
      copy: bool = False
  ) -> Union['Content', None]:
    """Returns a Content object or None from a writable type."""
    if value is None:
      return None

    if isinstance(value, Content):
      assert isinstance(value, cls), (value, cls)
      if copy:
        return copy_lib.deepcopy(value)
      return value
    return cls(value)   # pytype: disable=not-instantiable

  @classmethod
  def _to_content(cls, value: WritableTypes) -> Union['Content', str, None]:
    """Returns a Content object or None from a writable type."""
    if callable(value):
      value = value()
    if value is None:
      return None
    assert isinstance(value, (str, cls)), value
    return value


def view(
    value: Any,
    *,
    name: Optional[str] = None,
    root_path: Optional[object_utils.KeyPath] = None,
    view_id: str = 'html-tree-view',
    **kwargs
) -> Content:
  """Views an object through generating content based on a specific view.

  Args:
    value: The value to view.
    name: The name of the value.
    root_path: The root path of the value.
    view_id: The ID of the view to use. See `pg.View.dir()` for all available
      view IDs.
    **kwargs: Additional keyword arguments passed to the view, wich
      will be used as the preset arguments for the View and Extension methods.

  Returns:
    The rendered `Content` object.
  """
  if isinstance(value, Content):
    return value

  with view_options(**kwargs) as options:
    view_object = View.create(view_id)
    return view_object.render(
        value, name=name, root_path=root_path or object_utils.KeyPath(),
        **options
    )


@contextlib.contextmanager
def view_options(**kwargs) -> Iterator[Dict[str, Any]]:
  """Context manager to inject rendering args to view.

  Example:

    with pg.view_options(enable_summary_tooltip=False):
      MyObject().to_html()

  Args:
    **kwargs: Keyword arguments for View.render method.

  Yields:
    The merged keyword arguments.
  """
  parent_options = object_utils.thread_local_peek(_TLS_KEY_VIEW_OPTIONS, {})
  # Deep merge the two dict.
  options = object_utils.merge([parent_options, kwargs])
  object_utils.thread_local_push(_TLS_KEY_VIEW_OPTIONS, options)
  try:
    yield options
  finally:
    object_utils.thread_local_pop(_TLS_KEY_VIEW_OPTIONS)


class View(metaclass=abc.ABCMeta):
  """Base class for views.

  A view defines an unique way/format of rendering an object (e.g. int, str,
  list, dict, user-defined object). For example, `pg.HtmlView` is a view that
  renders an object into HTML. `pg.HtmlTreeViews` is a concrete `pg.HtmlView`
  that renders an object into a HTML tree view.
  """

  # The ID of the view, which will be used as the value for the `view`
  # argument of `pg.view()`. Must be set for non-abstract subclasses.
  VIEW_ID = None

  class Extension:
    """Extension for the View class.

    View developers should always create a corresponding ``Extension`` class as
    an inner class of the ``View`` class. The ``Extension`` class defines custom
    rendering logic for user-defined types and provides methods that can be
    bound to the ``View`` methods via the ``@pg.View.extension_method``
    decorator.

    Example:

    .. code-block:: python

      class MyView(pg.View):
        VIEW_TYPE = 'my_view'

        class Extension(pg.View.Extension):

          def _my_view_render(self, value, *, view, **kwargs):
            return view.render(value, **kwargs)

          def _my_view_title(self, value, *, view, **kwargs):
            return view.render_title(value, **kwargs)

        @pg.View.extension_method('_my_view_title')
        def title(self, value: Any, **kwargs):
          return pg.Html('<h1>' + str(value) + '</h1>')

        @pg.View.extension_method('_my_view_render')
        def render(self, value: Any, **kwargs):
          return pg.Html(str(value))

    To use the ``Extension`` class, users can subclass it and override the
    extension methods in their own classes.

    For example:

    .. code-block:: python

      class MyObject(MyView.Extension):

        def _my_view_title(self, value, *, view, **kwargs):
          return pg.Html('Custom title for ' + str(value) + '</h1>')

        def _my_view_render(self, value, *, view, **kwargs):
          return self._my_view_title(value, **kwargs) + pg.Html(str(value))

    In this example, ``MyObject`` subclasses the ``Extension`` class of
    ``MyView`` and overrides the ``_my_view_title`` and ``_my_view_render``
    methods to provide custom view rendering for the object.
    """

    @classmethod
    @functools.cache
    def supported_view_classes(cls) -> Set[Type['View']]:
      """Returns all non-abstract View classes that the current class supports.

      A class can inherit from multiple ``View.Extension`` classes. For example:

      .. code-block:: python

        class MyObject(View1.Extension, View2.Extension):
          ...

      In this case, ``MyObject`` supports both ``View1`` and ``View2``.

      Returns:
        All non-abstract View classes that the current class supports.
      """
      supported_view_classes = set()
      view_class = pg_typing.get_outer_class(
          cls, base_cls=View, immediate=True
      )
      if view_class is not None and not inspect.isabstract(view_class):
        supported_view_classes.add(view_class)

      for base_cls in cls.__bases__:
        if issubclass(base_cls, View.Extension):
          supported_view_classes.update(base_cls.supported_view_classes())
      return supported_view_classes

  @classmethod
  def extension_method(cls, method_name: str) -> Any:
    """Decorator that dispatches a View method to a View.Extension method.

    A few things to note:
    1) The View method being decorated must have a `value` argument, based on
       which the Extension method will be dispatched.
    2) The View method's `value` argument will map to the Extension method's
       `self` argument.
    3) The Extension method can optionally have a `view` argument, which will
       be set to the current View class.

    Args:
      method_name: The name of the method in the Extension class to dispatch
        from current View method.

    Returns:
      A decorator that dispatches a View method to a View.Extension method.
    """

    def decorator(func):
      sig = pg_typing.signature(
          func, auto_typing=False, auto_doc=False
      )
      # We substract 1 to offset the `self` argument.
      try:
        extension_arg_index = sig.arg_names.index('value') - 1
      except ValueError as e:
        raise TypeError(
            f'View method {func.__name__!r} must have a `value` argument, '
            'which represents the target object to render.'
        ) from e
      if sig.varargs is not None:
        raise TypeError(
            f'View method must not have variable positional argument. '
            f'Found `*{sig.varargs.name}` in {func.__name__!r}'
        )

      def get_extension(args: Sequence[Any], kwargs: Dict[str, Any]) -> Any:
        if 'value' in kwargs:
          return kwargs['value']
        if extension_arg_index < len(args):
          return args[extension_arg_index]
        raise ValueError(
            'No value is provided for the `value` argument '
            f'for {func.__name__!r}.'
        )

      def map_args(
          args: Sequence[Any], kwargs: Dict[str, Any]
      ) -> Dict[str, Any]:
        # `args` does not contain self, therefore we use less than.
        assert len(args) < len(sig.args), (args, sig.args)
        kwargs.update({
            sig.args[i].name: arg
            for i, arg in enumerate(args) if i != extension_arg_index
        })
        kwargs.pop('value', None)
        return kwargs

      # We use the original function signature to generate the view method.
      @functools.wraps(func)
      def _generated_view_fn(self, *args, **kwargs):
        # This allows a View method to consume the preset kwargs from the
        # parent call to `pg.view`, yet also customizes the preset kwargs
        # to the calls to other View/Extension methods within this context.
        return self._maybe_dispatch(   # pylint: disable=protected-access
            *args, **kwargs,
            extension=get_extension(args, kwargs),
            view_method=func,
            extension_method_name=method_name,
            arg_map_fn=map_args
        )
      return _generated_view_fn
    return decorator

  def __init_subclass__(cls):
    if inspect.isabstract(cls):
      return

    # Register the view type.
    if cls.VIEW_ID is None:
      raise ValueError(
          f'`VIEW_ID` must be set for non-abstract View subclass {cls!r}.'
      )
    if cls.VIEW_ID == cls.__base__.VIEW_ID:
      raise ValueError(
          f'The `VIEW_ID` {cls.VIEW_ID!r} is the same as the base class '
          f'{cls.__base__.__name__}. Please choose a different ID.'
      )

    _VIEW_REGISTRY[cls.VIEW_ID] = cls
    super().__init_subclass__()

  @classmethod
  def dir(cls) -> Dict[str, Type['View']]:
    """Returns all registered View classes with their view IDs."""
    return {
        view_id: view_cls
        for view_id, view_cls in _VIEW_REGISTRY.items()
        if issubclass(view_cls, cls)
    }

  @staticmethod
  def create(view_id: str, **kwargs) -> 'View':
    """Creates a View instance with the given view ID."""
    if view_id not in _VIEW_REGISTRY:
      raise ValueError(
          f'No view class found with VIEW_ID: {view_id!r}'
      )
    return _VIEW_REGISTRY[view_id](**kwargs)

  def __init__(self, **kwargs):
    del kwargs
    super().__init__()

  @abc.abstractmethod
  def render(
      self,
      value: Any,
      *,
      name: Optional[str] = None,
      root_path: Optional[object_utils.KeyPath] = None,
      **kwargs
  ) -> Content:
    """Renders the input value.

    Args:
      value: The value to render.
      name: (Optional) The referred name of the value from its container.
      root_path: (Optional) The path of `value` under its object tree.
      **kwargs: Additional keyword arguments passed from `pg.view` or wrapper
        functions (e.g. `pg.to_html`).

    Returns:
      The rendered content.
    """

  #
  # Implementation for routing calls to View methods to to Extension methods.
  #

  def _maybe_dispatch(
      self,
      *args,
      extension: Any,
      view_method: types.FunctionType,
      extension_method_name: str,
      arg_map_fn: Callable[[Sequence[Any], Dict[str, Any]], Dict[str, Any]],
      **kwargs
  ) -> Any:
    """Dispatches the call to View method to corresponding Extension method.

    The dispatching should take care of these three scenarios.

    Scenario 1:
      The user code within an Extension delegates the handling of value `v` back
      to a View method.

      In this case, we should dispatch the call on `v` to the Extension method
      first, and use the View method if we detects that the request is routed
      back to the View method based on the same value.

    Scenario 2:
      The Extension (user code) employes a View method to handle a value `v`'s
      child value `w`.

      In this case, we should allow the Extension method to handle `w` first.

    Scenario 3:
      A Extension (custom render) uses another method of the view to render the
      same value `v`:

    To address these 3 scenarios, we use a thread-local stack to track the
    value being rendered by each view method.

    Args:
      *args: Positional arguments passed to the render method.
      extension: Extension to render.
      view_method: The default render method to call.
      extension_method_name: The name of the method in the Extension class
        to call.
      arg_map_fn: A function to map the view method args to the extension method
        args.
      **kwargs: Keyword arguments passed to the render method.

    Returns:
      The rendered HTML.
    """
    # No Extension involved, call the view's default render method.
    if not isinstance(extension, self.Extension):
      return view_method(self, *args, **kwargs)

    # Identify whether an extension is calling the view's method
    # within its extension method, in such case, we should delegate the
    # rendering logic to the default render method to avoid infinite recursion.
    with self._track_rendering(extension, view_method) as being_rendered:
      if being_rendered is extension:
        return view_method(self, *args, **kwargs)

      # Call the extension's method.
      mapped = arg_map_fn(args, kwargs)
      return getattr(extension, extension_method_name)(
          view=self, **mapped
      )

  @contextlib.contextmanager
  def _track_rendering(
      self,
      value: Extension,
      view_method
  ) -> Iterator[Any]:
    """Context manager for tracking the value being rendered."""
    del self
    rendering_stack = object_utils.thread_local_get(
        _TLS_KEY_OPERAND_STACK_BY_METHOD, {}
    )
    callsite_value = rendering_stack.get(view_method, None)
    rendering_stack[view_method] = value
    object_utils.thread_local_set(
        _TLS_KEY_OPERAND_STACK_BY_METHOD, rendering_stack
    )
    try:
      yield callsite_value
    finally:
      if callsite_value is None:
        rendering_stack.pop(view_method)
        if not rendering_stack:
          object_utils.thread_local_del(_TLS_KEY_OPERAND_STACK_BY_METHOD)
      else:
        rendering_stack[view_method] = callsite_value
